# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

from __future__ import annotations

__all__ = [
    'Token',
    'Tokenization',
    'Tokenizer',
    'RegexTokenizer',
    'SplittingTokenizer',
    'CharacterTokenizer',
    'WordTokenizer',
    'SentenceTokenizer',
]

from abc import ABC, abstractmethod
from dataclasses import dataclass
import icu
import threading
from typing import Callable, Iterable, Sequence

from ._alignment import Alignment
from ._bistr import bistr
from ._regex import compile_regex
from ._typing import Bounds, Regex, String


@dataclass(frozen=True)
class Token:
    """
    A token extracted from a string.
    """

    text: bistr
    """
    The actual text of the token.
    """

    start: int
    """
    The start position of the token.
    """

    end: int
    """
    The end position of the token.
    """

    @property
    def original(self) -> str:
        """
        The original value of this token.
        """
        return self.text.original

    @property
    def modified(self) -> str:
        """
        The modified value of this token.
        """
        return self.text.modified

    @classmethod
    def slice(cls, text: bistr, start: int, end: int) -> Token:
        """
        Create a Token from a slice of a bistr.
        """
        return cls(text[start:end], start, end)

    def __str__(self):
        return f'[{self.start}:{self.end}]={self.text}'

    def __repr__(self):
        return f'Token({self.text!r}, start={self.start}, end={self.end})'


@dataclass(frozen=True)
class Tokenization:
    """
    A string and its tokenization.
    """

    text: bistr
    """
    The text that was tokenized.
    """

    alignment: Alignment
    """
    The alignment from text indices to token indices.
    """

    _tokens: Sequence[Token]

    def __init__(self, text: bistr, tokens: Iterable[Token]):
        """
        Create a Tokenization.
        """
        tokens = tuple(tokens)

        alignment = []
        for i, token in enumerate(tokens):
            alignment.append((token.start, i))
            alignment.append((token.end, i + 1))

        super().__setattr__('text', text)
        super().__setattr__('_tokens', tokens)
        super().__setattr__('alignment', Alignment(alignment))

    def __iter__(self):
        return iter(self._tokens)

    def __len__(self):
        return len(self._tokens)

    def __getitem__(self, index):
        if isinstance(index, slice):
            start, stop, stride = index.indices(len(self))
            if stride != 1:
                raise ValueError('Non-unit strides not supported')

            text = self.substring(start, stop)
            tokens = self._tokens[index]
            if tokens:
                delta = tokens[0].start
                tokens = [Token(t.text, t.start - delta, t.end - delta) for t in tokens]
            return Tokenization(text, tokens)
        else:
            return self._tokens[index]

    def __str__(self):
        tokens = ', '.join(map(str, self))
        return f'Tokenization({self.text}, [{tokens}])'

    def __repr__(self):
        return f'Tokenization({self.text!r}, {self._tokens!r})'

    def substring(self, *args) -> Bounds:
        """
        Map a span of tokens to the corresponding substring.
        """
        return self.text[self.alignment.original_slice(*args)]

    def text_bounds(self, *args) -> Bounds:
        """
        Map a span of tokens to the bounds of the corresponding text.
        """
        return self.alignment.original_bounds(*args)

    def original_bounds(self, *args) -> Bounds:
        """
        Map a span of tokens to the bounds of the corresponding original text.
        """
        return self.text.alignment.original_bounds(self.text_bounds(*args))

    def bounds_for_text(self, *args) -> Bounds:
        """
        Map a span of text to the bounds of the corresponding span of tokens.
        """
        return self.alignment.modified_bounds(*args)

    def bounds_for_original(self, *args) -> Bounds:
        """
        Map a span of original text to the bounds of the corresponding span of
        tokens.
        """
        text_bounds = self.text.alignment.modified_bounds(*args)
        return self.alignment.modified_bounds(text_bounds)

    def slice_by_text(self, *args) -> Tokenization:
        """
        Map a span of text to the corresponding span of tokens.
        """
        i, j = self.bounds_for_text(*args)
        return self[i:j]

    def slice_by_original(self, *args) -> Tokenization:
        """
        Map a span of the original text to the corresponding span of tokens.
        """
        i, j = self.bounds_for_original(*args)
        return self[i:j]

    def snap_text_bounds(self, *args) -> Bounds:
        """
        Expand a span of text to align it with token boundaries.
        """
        return self.text_bounds(self.bounds_for_text(*args))

    def snap_original_bounds(self, *args) -> Bounds:
        """
        Expand a span of original text to align it with token boundaries.
        """
        return self.original_bounds(self.bounds_for_original(*args))


class Tokenizer(ABC):
    """
    Abstract base class for tokenizers.
    """

    @abstractmethod
    def tokenize(self, text: String) -> Tokenization:
        """
        Tokenize some text.

        :param text: The text to tokenize, as either an `str` or
                     :class:`~bistring.bistr`.  A plain `str` should be
                     converted to a `bistr` before processing.

        :returns: A :class:`~bistring.Tokenization` holding the text and its
                  tokens.
        """

        pass


class RegexTokenizer(Tokenizer):
    """
    Breaks text into tokens based on a regex.
    """

    def __init__(self, regex: Regex):
        self._pattern = compile_regex(regex)

    def tokenize(self, text: String) -> Tokenization:
        text = bistr(text)
        tokens = []
        for match in self._pattern.finditer(text.modified):
            tokens.append(Token.slice(text, match.start(), match.end()))
        return Tokenization(text, tokens)


class SplittingTokenizer(Tokenizer):
    """
    Splits text into tokens based on a regex.
    """

    def __init__(self, regex: Regex):
        self._pattern = compile_regex(regex)

    def tokenize(self, text: String) -> Tokenization:
        text = bistr(text)
        tokens = []

        last = 0
        for match in self._pattern.finditer(text.modified):
            start = match.start()
            if start > last:
                tokens.append(Token.slice(text, last, start))
            last = match.end()

        end = len(text.modified)
        if end > last:
            tokens.append(Token.slice(text, last, end))

        return Tokenization(text, tokens)


class _IcuTokenizer(Tokenizer):
    """
    Base class for ICU BreakIterator-based tokenizers.
    """

    def __init__(self, locale: str, constructor: Callable):
        # BreakIterator is not a thread-safe API, so store a cache of
        # thread-local iterators
        self._locale = icu.Locale(locale)
        self._constructor = constructor
        self._local = threading.local()

        # Eagerly construct one on this thread as an optimization, and to check
        # for errors
        self._break_iterator()

    def _break_iterator(self) -> icu.BreakIterator:
        if not hasattr(self._local, 'bi'):
            self._local.bi = self._constructor(self._locale)
        return self._local.bi

    def tokenize(self, text: String) -> Tokenization:
        text = bistr(text)
        tokens = []

        bi = self._break_iterator()

        utext = icu.UnicodeString(text.modified)
        bi.setText(utext)

        ui = bi.first()
        uj = bi.nextBoundary()
        i = 0
        while uj != icu.BreakIterator.DONE:
            j = i + utext.countChar32(ui, uj - ui)
            if self._check_token(bi.getRuleStatus()):
                tokens.append(Token.slice(text, i, j))
            ui = uj
            uj = bi.nextBoundary()
            i = j

        return Tokenization(text, tokens)

    def _check_token(self, tag: int) -> bool:
        return True


class CharacterTokenizer(_IcuTokenizer):
    """
    Splits text into user-perceived characters/grapheme clusters.
    """

    def __init__(self, locale: str):
        super().__init__(locale, icu.BreakIterator.createCharacterInstance)


class WordTokenizer(_IcuTokenizer):
    """
    Splits text into words based on Unicode rules.
    """

    def __init__(self, locale: str):
        super().__init__(locale, icu.BreakIterator.createWordInstance)

    def _check_token(self, tag: int) -> bool:
        return tag >= 100 # UBRK_WORD_NONE_LIMIT


class SentenceTokenizer(_IcuTokenizer):
    """
    Splits text into sentences based on Unicode rules.
    """

    def __init__(self, locale: str):
        super().__init__(locale, icu.BreakIterator.createSentenceInstance)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

__all__ = ['BistrBuilder']

from typing import Iterable, List, Match, Optional, Pattern, Tuple

from ._alignment import Alignment
from ._bistr import bistr
from ._regex import compile_regex, expand_template
from ._typing import Bounds, Regex, Replacement, String


class BistrBuilder:
    """
    Bidirectionally transformed string builer.
    """

    _original: bistr
    _modified: List[str]
    _alignment: List[Bounds]
    _opos: int
    _mpos: int

    def __init__(self, original: String):
        self._original = bistr(original)
        self._modified = []
        self._alignment = [(0, 0)]
        self._opos = 0
        self._mpos = 0

    @property
    def original(self) -> str:
        """
        The original string being modified.
        """
        return self._original.original

    @property
    def current(self) -> str:
        """
        The current string before modifications.
        """
        return self._original.modified

    @property
    def modified(self) -> str:
        """
        The modified string as built so far.
        """
        return ''.join(self._modified)

    @property
    def alignment(self) -> Alignment:
        """
        The alignment built so far from self.current to self.modified.
        """
        return Alignment(self._alignment)

    @property
    def position(self) -> int:
        """
        The position of the builder in self.current.
        """
        return self._opos

    @property
    def remaining(self) -> int:
        """
        The number of characters of the current string left to process.
        """
        return len(self.current) - self._opos

    @property
    def is_complete(self) -> bool:
        """
        Whether we've completely processed the string.
        """
        return self.remaining == 0

    def peek(self, n: int):
        """
        Peek at the next n characters of the original string.
        """
        return self.current[self._opos:self._opos+n]

    def _advance(self, ocount, mcount):
        self._opos += ocount
        self._mpos += mcount
        if ocount > 0 or mcount > 0:
            self._alignment.append((self._opos, self._mpos))

    def skip(self, n: int):
        """
        Skip the next n characters, copying them unchanged.
        """
        if n > 0:
            self._modified.append(self.peek(n))
            for i in range(n):
                self._advance(1, 1)

    def skip_rest(self):
        """
        Skip the rest of the string, copying it unchanged.
        """
        self.skip(self.remaining)

    def insert(self, string: str):
        """
        Insert a substring into the string.
        """
        self.replace(0, string)

    def discard(self, n: int):
        """
        Discard a portion of the original string.
        """
        self.replace(n, '')

    def discard_rest(self):
        """
        Discard the rest of the original string.
        """
        self.discard(self.remaining)

    def replace(self, n: int, repl: str):
        """
        Replace the next n characters with a new string.
        """
        if len(repl) > 0:
            self._modified.append(repl)
        self._advance(n, len(repl))

    def append(self, bs: bistr):
        """
        Append a bistr.  The original value of the bistr must match the current
        string being processed.
        """
        if bs.original != self.peek(len(bs.original)):
            raise ValueError("bistr doesn't match the current string")
        for x, y in zip(bs.alignment, bs.alignment[1:]):
            self._advance(y[0] - x[0], y[1] - x[1])

    def _match(self, regex: Regex) -> Optional[Match[str]]:
        pattern = compile_regex(regex)
        return pattern.match(self.current, pos=self._opos)

    def _search(self, regex: Regex) -> Optional[Match[str]]:
        pattern = compile_regex(regex)
        return pattern.search(self.current, pos=self._opos)

    def _finditer(self, regex: Regex) -> Iterable[Match[str]]:
        pattern = compile_regex(regex)
        return pattern.finditer(self.current, pos=self._opos)

    def skip_match(self, regex: Regex) -> bool:
        """
        Skip a substring matching a regex, copying it unchanged.
        """
        match = self._match(regex)
        if match:
            self.skip(match.end() - match.start())
            return True
        else:
            return False

    def discard_match(self, regex: Regex) -> bool:
        """
        Discard a substring that matches a regex.
        """
        match = self._match(regex)
        if match:
            self.discard(match.end() - match.start())
            return True
        else:
            return False

    def replace_match(self, regex: Regex, repl: Replacement) -> bool:
        """
        Replace a substring that matches a regex.
        """
        match = self._match(regex)
        if match:
            self.replace(match.end() - match.start(), expand_template(match, repl))
            return True
        else:
            return False

    def replace_next(self, regex: Regex, repl: Replacement) -> bool:
        """
        Replace the next occurence of a regex.
        """
        match = self._search(regex)
        if match:
            self.skip(match.start() - self._opos)
            self.replace(match.end() - match.start(), expand_template(match, repl))
            return True
        else:
            return False

    def replace_all(self, regex: Regex, repl: Replacement):
        """
        Replace all occurences of a regex.
        """
        for match in self._finditer(regex):
            self.skip(match.start() - self._opos)
            self.replace(match.end() - match.start(), expand_template(match, repl))
        self.skip_rest()

    def build(self):
        """
        Build the bistr.
        """
        alignment = self._original.alignment.compose(self.alignment)
        return bistr(self.original, self.modified, alignment)

    def rewind(self):
        """
        Reset this builder to apply another transformation.
        """
        self._original = self.build()
        self._modified = []
        self._alignment = [(0, 0)]
        self._opos = 0
        self._mpos = 0



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

from __future__ import annotations

__all__ = ['Alignment']

import bisect
from typing import Iterable, List, Optional, Tuple, cast, overload

from ._typing import Bounds, Range


class Alignment:
    """
    An alignment between two related sequences.
    """

    __slots__ = ('_original', '_modified')

    _original: List[int]
    _modified: List[int]

    def __init__(self, values: Iterable[Bounds]):
        self._original = []
        self._modified = []
        for i, j in values:
            if self._original:
                if i < self._original[-1]:
                    raise ValueError('Original sequence position moved backwards')
                elif j < self._modified[-1]:
                    raise ValueError('Modified sequence position moved backwards')
                elif i == self._original[-1] and j == self._modified[-1]:
                    continue

            self._original.append(i)
            self._modified.append(j)

        if not self._original:
            raise ValueError('No sequence positions to align')

    @classmethod
    def _create(cls, original: List[int], modified: List[int]) -> Alignment:
        result = super().__new__(cls)
        result._original = original
        result._modified = modified
        return result

    def __str__(self):
        i, j = self._original[0], self._original[-1]
        k, l = self._modified[0], self._modified[-1]
        if self._original == list(range(i, j + 1)) and self._modified == list(range(k, l + 1)):
            return f'[{i}:{j}⇋{k}:{l}]'
        else:
            return '[' + ', '.join(f'{i}⇋{j}' for i, j in self) + ']'

    def __repr__(self):
        i, j = self._original[0], self._original[-1]
        if self._original == list(range(i, j + 1)) and self._modified == list(range(i, j + 1)):
            if i == 0:
                return f'Alignment.identity({j})'
            else:
                return f'Alignment.identity({i}, {j})'
        else:
            return 'Alignment([' + ', '.join(map(repr, self)) + '])'

    def __eq__(self, other):
        if isinstance(other, Alignment):
            return (self._original, self._modified) == (other._original, other._modified)
        else:
            return NotImplemented

    @classmethod
    def _parse_args(cls, args: Tuple) -> Bounds:
        l = len(args)
        if l == 0:
            return None, None
        elif l == 1:
            arg = args[0]
            if isinstance(arg, range):
                return arg.start, arg.stop
            elif isinstance(arg, slice):
                if arg.start is None or arg.stop is None:
                    raise ValueError('slice with unspecified bounds')
                return arg.start, arg.stop
            elif isinstance(arg, tuple):
                return cast(Bounds, arg)
            else:
                return 0, arg
        elif l == 2:
            return cast(Bounds, args)
        else:
            raise TypeError('Too many arguments')

    @overload
    @classmethod
    def identity(cls, length: int) -> Alignment:
        ...

    @overload
    @classmethod
    def identity(cls, start: int, stop: int) -> Alignment:
        ...

    @overload
    @classmethod
    def identity(cls, bounds: Range) -> Alignment:
        ...

    @classmethod
    def identity(cls, *args):
        start, stop = cls._parse_args(args)
        values = list(range(start, stop + 1))
        return cls._create(values, values)

    def __iter__(self):
        return zip(self._original, self._modified)

    def __len__(self):
        return len(self._original)

    def __getitem__(self, index):
        if isinstance(index, slice):
            start, stop, stride = index.indices(len(self))
            if stride != 1:
                raise ValueError('Non-unit strides not supported')
            return self._create(self._original[index], self._modified[index])
        else:
            return (self._original[index], self._modified[index])

    def shift(self, delta_o: int, delta_m: int):
        """
        Shift this alignment.

        :param delta_o: The distance to shift the original sequence.
        :param delta_m: The distance to shift the modified sequence.
        """

        return self._create(
            [o + delta_o for o in self._original],
            [m + delta_m for m in self._modified],
        )

    def _search(self, source: List[int], start: int, stop: int) -> Bounds:
        first = bisect.bisect_right(source, start)
        if first == 0:
            raise IndexError('range start too small')
        first -= 1

        last = bisect.bisect_left(source, stop, first)
        if last == len(source):
            raise IndexError('range end too big')

        return first, last

    def _bounds(self, source: List[int], target: List[int], args: Tuple) -> Bounds:
        start, stop = self._parse_args(args)
        if start is None:
            i, j = 0, -1
        else:
            i, j = self._search(source, start, stop)
        return (target[i], target[j])

    def original_bounds(self, *args) -> Bounds:
        return self._bounds(self._modified, self._original, args)

    def original_range(self, *args) -> range:
        return range(*self.original_bounds(*args))

    def original_slice(self, *args) -> slice:
        return slice(*self.original_bounds(*args))

    def modified_bounds(self, *args) -> Bounds:
        return self._bounds(self._original, self._modified, args)

    def modified_range(self, *args) -> range:
        return range(*self.modified_bounds(*args))

    def modified_slice(self, *args) -> slice:
        return slice(*self.modified_bounds(*args))

    def slice_by_original(self, *args) -> Alignment:
        start, stop = self._parse_args(args)
        first, last = self._search(self._original, start, stop)
        original = self._original[first:last+1]
        original = [min(max(i, start), stop) for i in original]
        modified = self._modified[first:last+1]
        return self._create(original, modified)

    def slice_by_modified(self, *args) -> Alignment:
        start, stop = self._parse_args(args)
        first, last = self._search(self._modified, start, stop)
        original = self._original[first:last+1]
        modified = self._modified[first:last+1]
        modified = [min(max(i, start), stop) for i in modified]
        return self._create(original, modified)

    def __add__(self, other):
        """
        Concatenate two alignments.
        """

        if not isinstance(other, Alignment):
            return NotImplemented

        o_orig = other._original
        o_mod = other._modified

        if o_orig[0] < self._original[-1]:
            raise ValueError('Original sequence position moved backwards')
        elif o_mod[0] < self._modified[-1]:
            raise ValueError('Modified sequence position moved backwards')
        elif o_orig[0] == self._original[-1] and o_mod[0] == self._modified[-1]:
            o_orig = o_orig[1:]
            o_mod = o_mod[1:]

        return self._create(self._original + o_orig, self._modified + o_mod)

    def compose(self, other: Alignment) -> Alignment:
        """
        Return a new alignment equivalent to applying this one first, then the
        other.
        """

        if self.modified_bounds() != other.original_bounds():
            raise ValueError('Incompatible alignments')

        original = []
        modified = []
        i, i_max = 0, len(self)
        j, j_max = 0, len(other)

        while i < i_max:
            # Map self._original[i] to its lower bound in other
            while self._modified[i] > other._original[j]:
                j += 1
            while self._modified[i] < other._original[j] and self._modified[i + 1] <= other._original[j]:
                i += 1
            original.append(self._original[i])
            modified.append(other._modified[j])

            # Map self._original[i] to its upper bound in other (if it's different)
            while i + 1 < i_max and self._original[i] == self._original[i + 1]:
                i += 1

            needs_upper = False
            while j + 1 < j_max and self._modified[i] >= other._original[j + 1]:
                needs_upper = True
                j += 1
            if needs_upper:
                original.append(self._original[i])
                modified.append(other._modified[j])

            i += 1

        return self._create(original, modified)

    def inverse(self) -> Alignment:
        """
        The inverse of this alignment, from the modified to the original sequence.
        """
        return self._create(self._modified, self._original)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

import re
from typing import Match, Pattern

from ._typing import Regex, Replacement


def compile_regex(regex: Regex) -> Pattern[str]:
    if isinstance(regex, str):
        return re.compile(regex)
    else:
        return regex


def expand_template(match: Match[str], repl: Replacement) -> str:
    if callable(repl):
        return repl(match)
    else:
        return match.expand(repl)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

from __future__ import annotations

__all__ = ['bistr']

from itertools import islice
from typing import Iterable, List, Optional, Tuple, Union

from ._alignment import Alignment
from ._typing import Bounds, Regex, Replacement, String


class bistr:
    """
    A bidirectionally transformed string.
    """

    __slots__ = ('original', 'modified', 'alignment')

    original: str
    """
    The original string, before any modifications.
    """

    modified: str
    """
    The current value of the string, after all modifications.
    """

    alignment: Alignment
    """
    The sequence alignment between :attr:`original` and :attr:`modified`.
    """

    def __new__(cls, original: String, modified: Optional[str] = None, alignment: Optional[Alignment] = None):
        """
        A `bistr` can be constructed from only a single string, which will give it identical original and modified
        strings and an identity alignment:

            >>> s = bistr('test')
            >>> s.original
            'test'
            >>> s.modified
            'test'
            >>> s.alignment
            Alignment.identity(4)

        You can also explicitly specify both the original and modified string.  The inferred alignment will be as course
        as possible:

            >>> s = bistr('TEST', 'test')
            >>> s.original
            'TEST'
            >>> s.modified
            'test'
            >>> s.alignment
            Alignment([(0, 0), (4, 4)])

        Finally, you can specify the alignment explicitly too, if you know it:

            >>> s = bistr('TEST', 'test', Alignment.identity(4))
            >>> s[1:3]
            bistr('ES', 'es', Alignment.identity(2))
        """

        if isinstance(original, bistr):
            if modified is not None or alignment is not None:
                raise ValueError('bistr copy constructor invoked with extra arguments')
            return original
        elif not isinstance(original, str):
            raise TypeError(f'Expected a string, found {type(original)}')

        if modified is None:
            modified = original
            if alignment is None:
                alignment = Alignment.identity(len(original))
        elif isinstance(modified, str):
            if alignment is None:
                alignment = Alignment([(0, 0), (len(original), len(modified))])
        else:
            raise TypeError(f'Expected a string, found {type(modified)}')

        if not isinstance(alignment, Alignment):
            raise TypeError(f'Expected an Alignment, found {type(alignment)}')

        if alignment.original_bounds() != (0, len(original)):
            raise ValueError('Alignment incompatible with original string')
        elif alignment.modified_bounds() != (0, len(modified)):
            raise ValueError('Alignment incompatible with modified string')

        result = super().__new__(cls)
        super().__setattr__(result, 'original', original)
        super().__setattr__(result, 'modified', modified)
        super().__setattr__(result, 'alignment', alignment)
        return result

    def __str__(self):
        if self.original == self.modified:
            return f'⮎{self.original!r}⮌'
        else:
            return f'({self.original!r} ⇋ {self.modified!r})'

    def __repr__(self):
        if self.original == self.modified and len(self.alignment) == len(self.original) + 1:
            return f'bistr({self.original!r})'
        elif len(self.alignment) == 2:
            return f'bistr({self.original!r}, {self.modified!r})'
        else:
            return f'bistr({self.original!r}, {self.modified!r}, {self.alignment!r})'

    def __len__(self):
        return len(self.modified)

    def __eq__(self, other):
        if isinstance(other, bistr):
            return (self.original, self.modified, self.alignment) == (other.original, other.modified, other.alignment)
        else:
            return NotImplemented

    def __add__(self, other):
        if isinstance(other, bistr):
            original = other.original
            modified = other.modified
            alignment = other.alignment
        elif isinstance(other, str):
            original = other
            modified = other
            alignment = Alignment.identity(len(other))
        else:
            return NotImplemented

        alignment = alignment.shift(len(self.original), len(self.modified))
        return bistr(self.original + original, self.modified + modified, self.alignment + alignment)

    def __radd__(self, other):
        if isinstance(other, str):
            length = len(other)
            return bistr(
                other + self.original,
                other + self.modified,
                Alignment.identity(length) + self.alignment.shift(length, length),
            )
        else:
            return NotImplemented

    def __getitem__(self, index):
        """
        Indexing a `bistr` returns the nth character of the modified string:

            >>> s = bistr('TEST').lower()
            >>> s[1]
            'e'

        Slicing a `bistr` extracts a substring, complete with the matching part
        of the original string:

            >>> s = bistr('TEST').lower()
            >>> s[1:3]
            bistr('ES', 'es', Alignment.identity(2))
        """

        if isinstance(index, slice):
            start, stop, stride = index.indices(len(self))
            if stride != 1:
                raise ValueError('Non-unit strides not supported')

            modified = self.modified[start:stop]
            original = self.original[self.alignment.original_slice(start, stop)]
            alignment = self.alignment.slice_by_modified(start, stop)
            alignment = alignment.shift(-alignment[0][0], -alignment[0][1])

            return bistr(original, modified, alignment)
        else:
            return self.modified[index]

    def __setattr__(self, name, value):
        raise AttributeError('bistr is immutable')

    def __delattr__(self, name):
        raise AttributeError('bistr is immutable')

    def inverse(self) -> bistr:
        """
        :returns: The inverse of this string, swapping the original and modified strings.

            >>> s = bistr('HELLO WORLD').lower()
            >>> s
            bistr('HELLO WORLD', 'hello world', Alignment.identity(11))
            >>> s.inverse()
            bistr('hello world', 'HELLO WORLD', Alignment.identity(11))
        """
        return bistr(self.modified, self.original, self.alignment.inverse())

    def chunks(self) -> Iterable[bistr]:
        """
        :returns: All the chunks of associated text in this string.
        """

        i, k = 0, 0
        for j, l in self.alignment[1:]:
            yield bistr(self.original[i:j], self.modified[k:l])
            i, k = j, l

    def count(self, sub: str, start: Optional[int] = None, end: Optional[int] = None) -> int:
        """
        Like :meth:`str.count`, counts the occurrences of `sub` in the string.
        """
        return self.modified.count(sub, start, end)

    def find(self, sub: str, start: Optional[int] = None, end: Optional[int] = None) -> int:
        """
        Like :meth:`str.find`, finds the position of `sub` in the string.
        """
        return self.modified.find(sub, start, end)

    def find_bounds(self, sub: str, start: Optional[int] = None, end: Optional[int] = None) -> Bounds:
        """
        Like :meth:`find`, but returns both the start and end bounds for convenience.

        :returns: The first `i, j` within `[start, end)` such that ``self[i:j] == sub``, or ``(-1, -1)`` if not found.
        """

        i = self.find(sub, start, end)
        if i >= 0:
            return i, i + len(sub)
        else:
            return i, i

    def rfind(self, sub: str, start: Optional[int] = None, end: Optional[int] = None) -> int:
        """
        Like :meth:`str.rfind`, finds the position of `sub` in the string backwards.
        """
        return self.modified.rfind(sub, start, end)

    def rfind_bounds(self, sub: str, start: Optional[int] = None, end: Optional[int] = None) -> Bounds:
        """
        Like :meth:`rfind`, but returns both the start and end bounds for convenience.

        :returns: The last `i, j` within `[start, end)` such that ``self[i:j] == sub``, or ``(-1, -1)`` if not found.
        """

        i = self.rfind(sub, start, end)
        if i >= 0:
            return i, i + len(sub)
        else:
            return i, i

    def index(self, sub: str, start: Optional[int] = None, end: Optional[int] = None) -> int:
        """
        Like :meth:`str.index`, finds the first position of `sub` in the string, otherwise raising a `ValueError`.
        """
        return self.modified.index(sub, start, end)

    def index_bounds(self, sub: str, start: Optional[int] = None, end: Optional[int] = None) -> Bounds:
        """
        Like :meth:`index`, but returns both the start and end bounds for convenience.  If the substring is not found, a
        :class:`ValueError` is raised.

        :returns: The first `i, j` within `[start, end)` such that ``self[i:j] == sub``.
        :raises: :class:`ValueError` if the substring is not found.
        """

        i = self.index(sub, start, end)
        return i, i + len(sub)

    def rindex(self, sub: str, start: Optional[int] = None, end: Optional[int] = None) -> int:
        """
        Like :meth:`str.index`, finds the last position of `sub` in the string, otherwise raising a `ValueError`.
        """
        return self.modified.rindex(sub, start, end)

    def rindex_bounds(self, sub: str, start: Optional[int] = None, end: Optional[int] = None) -> Bounds:
        """
        Like :meth:`rindex`, but returns both the start and end bounds for convenience.  If the substring is not found, a
        :class:`ValueError` is raised.

        :returns: The last `i, j` within `[start, end)` such that ``self[i:j] == sub``.
        :raises: :class:`ValueError` if the substring is not found.
        """

        i = self.rindex(sub, start, end)
        return i, i + len(sub)

    def startswith(self, prefix: Union[str, Tuple[str, ...]], start: Optional[int] = None, end: Optional[int] = None) -> bool:
        """
        Like :meth:`str.startswith`, checks if the string starts with the given `prefix`.
        """
        return self.modified.startswith(prefix, start, end)

    def endswith(self, suffix: Union[str, Tuple[str, ...]], start: Optional[int] = None, end: Optional[int] = None) -> bool:
        """
        Like :meth:`str.endswith`, checks if the string starts with the given `suffix`.
        """
        return self.modified.endswith(suffix, start, end)

    @classmethod
    def join(cls, iterable: Iterable[String]) -> bistr:
        """
        Like :meth:`str.join`, concatenates many (bi)strings together.
        """

        result = cls('')
        for element in iterable:
            result += cls(element)
        return result

    def _find_whitespace(self, start: int) -> Bounds:
        for i in range(start, len(self)):
            if self[i].isspace():
                first = i
                break
        else:
            return -1, -1

        for i in range(first + 1, len(self)):
            if not self[i].isspace():
                last = i
                break
        else:
            last = len(self)

        return first, last

    def split(self, sep: Optional[str] = None, maxsplit: int = -1) -> List[bistr]:
        """
        Like :meth:`str.split`, splits this string on a separator.
        """

        result = []
        count = 0
        start = 0

        while start >= 0 and (count < maxsplit or maxsplit == -1):
            if sep is None:
                i, j = self._find_whitespace(start)
            else:
                i, j = self.find_bounds(sep, start)

            if i < 0:
                i = len(self)

            if i > start or sep is not None:
                result.append(self[start:i])
                count += 1

            start = j

        if start >= 0:
            result.append(self[start:])

        return result

    def partition(self, sep: str) -> Tuple[bistr, bistr, bistr]:
        """
        Like :meth:`str.partition`, splits this string into three chunks on a separator.
        """

        i, j = self.find_bounds(sep)
        if i >= 0:
            return self[:i], self[i:j], self[j:]
        else:
            return self, bistr(), bistr()

    def rpartition(self, sep: str) -> Tuple[bistr, bistr, bistr]:
        """
        Like :meth:`str.rpartition`, splits this string into three chunks on a separator, searching from the end.
        """

        i, j = self.rfind_bounds(sep)
        if i >= 0:
            return self[:i], self[i:j], self[j:]
        else:
            return self, bistr(), bistr()

    def center(self, width: int, fillchar: str = ' ') -> bistr:
        """
        Like :meth:`str.center`, pads the start and end of the string to center it.
        """

        if len(self) >= width:
            return self

        pad = width - len(self)
        lpad = pad // 2
        rpad = (pad + 1) // 2
        return bistr('', fillchar * lpad) + self + bistr('', fillchar * rpad)

    def ljust(self, width: int, fillchar: str = ' ') -> bistr:
        """
        Like :meth:`str.ljust`, pads the end of the string to a fixed width.
        """

        if len(self) >= width:
            return self

        pad = width - len(self)
        return self + bistr('', fillchar * pad)

    def rjust(self, width: int, fillchar: str = ' ') -> bistr:
        """
        Like :meth:`str.rjust`, pads the start of the string to a fixed width.
        """

        if len(self) >= width:
            return self

        pad = width - len(self)
        return bistr('', fillchar * pad) + self

    def _builder(self):
        from ._builder import BistrBuilder
        return BistrBuilder(self)

    def casefold(self) -> bistr:
        """
        Computes the case folded form of this string.  Case folding is used for case-insensitive operations, and the
        result may not be suitable for displaying to a user.  For example:

            >>> s = bistr('straße').casefold()
            >>> s.modified
            'strasse'
            >>> s[4:6]
            bistr('ß', 'ss')
        """

        from ._icu import casefold
        return casefold(self)

    def lower(self, locale: Optional[str] = None) -> bistr:
        """
        Converts this string to lowercase.  Unless you specify the `locale` parameter, the current system locale will be
        used.

        >>> bistr('HELLO WORLD').lower()
        bistr('HELLO WORLD', 'hello world', Alignment.identity(11))
        >>> bistr('I').lower('en_US')
        bistr('I', 'i')
        >>> bistr('I').lower('tr_TR')
        bistr('I', 'ı')
        """

        from ._icu import lower
        return lower(self, locale)

    def upper(self, locale: Optional[str] = None) -> bistr:
        """
        Converts this string to uppercase.  Unless you specify the `locale` parameter, the current system locale will be
        used.

        >>> bistr('hello world').upper()
        bistr('hello world', 'HELLO WORLD', Alignment.identity(11))
        >>> bistr('i').upper('en_US')
        bistr('i', 'I')
        >>> bistr('i').upper('tr_TR')
        bistr('i', 'İ')
        """

        from ._icu import upper
        return upper(self, locale)

    def title(self, locale: Optional[str] = None) -> bistr:
        """
        Converts this string to title case.  Unless you specify the `locale` parameter, the current system locale will
        be used.

        >>> bistr('hello world').title()
        bistr('hello world', 'Hello World', Alignment.identity(11))
        >>> bistr('istanbul').title('en_US')
        bistr('istanbul', 'Istanbul', Alignment.identity(8))
        >>> bistr('istanbul').title('tr_TR')
        bistr('istanbul', 'İstanbul', Alignment.identity(8))
        """

        from ._icu import title
        return title(self, locale)

    def capitalize(self, locale: Optional[str] = None) -> bistr:
        """
        Capitalize the first character of this string, and lowercase the rest.  Unless you specify the `locale`
        parameter, the current system locale will be used.

            >>> bistr('hello WORLD').capitalize()
            bistr('hello WORLD', 'Hello world', Alignment.identity(11))
            >>> bistr('ἴΣ').capitalize('el_GR')
            bistr('ἴΣ', 'Ἴς', Alignment.identity(2))
        """

        # We have to be careful here to get context-sensitive letters like
        # word-final sigma correct

        builder = self._builder()

        title = bistr(self.modified).title()
        for chunk in islice(title.chunks(), 1):
            builder.replace(len(chunk.original), chunk.modified)

        lower = bistr(self.modified).lower()
        for chunk in islice(lower.chunks(), 1, None):
            builder.replace(len(chunk.original), chunk.modified)

        return builder.build()

    def expandtabs(self, tabsize: int = 8) -> bistr:
        """
        Like :meth:`str.expandtabs`, replaces tab (``\\t``) characters with spaces to align on multiples of `tabsize`.
        """

        builder = self._builder()

        col = 0
        while not builder.is_complete:
            c = builder.peek(1)
            if c == '\t':
                spaces = tabsize - (col % tabsize)
                builder.replace(1, ' ' * spaces)
                col += spaces
            else:
                builder.skip(1)
                if c == '\n' or c == '\r':
                    col = 0
                else:
                    col += 1

        return builder.build()

    def replace(self, old: str, new: str, count: Optional[int] = None) -> bistr:
        """
        Like :meth:`str.replace`, replaces occurrences of `old` with `new`.
        """

        builder = self._builder()

        pos = 0
        n = 0
        while count is None or n < count:
            index = self.modified.find(old, pos)
            if index < 0:
                break

            builder.skip(index - pos)
            builder.replace(len(old), new)

            pos = index + len(old)
            n += 1

        builder.skip_rest()
        return builder.build()

    def sub(self, regex: Regex, repl: Replacement) -> bistr:
        """
        Like :meth:`re.sub`, replaces all matches of `regex` with the replacement `repl`.

        :param regex:
            The regex to match.  Can be a string pattern or a compiled regex.
        :param repl:
            The replacement to use.  Can be a string, which is interpreted as in :meth:`re.Match.expand`, or a
            `callable`, which will receive each match and return the replacement string.
        """

        builder = self._builder()
        builder.replace_all(regex, repl)
        return builder.build()

    def _stripper(self, chars: Optional[str]):
        if chars is None:
            return lambda c: c.isspace()
        else:
            return lambda c: c in chars

    def strip(self, chars: Optional[str] = None) -> bistr:
        """
        Like :meth:`str.strip`, removes leading and trailing characters (whitespace by default).
        """

        should_strip = self._stripper(chars)

        length = len(self)
        pre = 0
        while pre < length and should_strip(self.modified[pre]):
            pre += 1

        post = length
        while post > pre and should_strip(self.modified[post - 1]):
            post -= 1

        builder = self._builder()
        builder.discard(pre)
        builder.skip(post - pre)
        builder.discard_rest()
        return builder.build()

    def lstrip(self, chars: Optional[str] = None) -> bistr:
        """
        Like :meth:`str.lstrip`, removes leading characters (whitespace by default).
        """

        should_strip = self._stripper(chars)

        length = len(self)
        pre = 0
        while pre < length and should_strip(self.modified[pre]):
            pre += 1

        builder = self._builder()
        builder.discard(pre)
        builder.skip_rest()
        return builder.build()

    def rstrip(self, chars: Optional[str] = None) -> bistr:
        """
        Like :meth:`str.rstrip`, removes trailing characters (whitespace by default).
        """

        should_strip = self._stripper(chars)

        length = len(self)
        post = length
        while post > 0 and should_strip(self.modified[post - 1]):
            post -= 1

        builder = self._builder()
        builder.skip(post)
        builder.discard_rest()
        return builder.build()

    def normalize(self, form: str) -> bistr:
        """
        Like :meth:`unicodedata.normalize`, applies a Unicode `normalization form <https://unicode.org/reports/tr15/#Norm_Forms>`_.
        The choices for `form` are:

        - ``'NFC'``: Canonical Composition
        - ``'NFKC'``: Compatibility Composition
        - ``'NFD'``: Canonical Decomposition
        - ``'NFKD'``: Compatibilty Decomposition
        """

        from ._icu import normalize
        return normalize(self, form)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

import icu
from typing import Callable, Optional

from ._bistr import bistr
from ._builder import BistrBuilder


def _edit(bs: bistr, op: Callable, locale: Optional[str] = None) -> bistr:
    builder = BistrBuilder(bs)
    edits = icu.Edits()
    ucur = icu.UnicodeString(builder.current)

    if locale is None:
        umod = icu.UnicodeString(op(ucur, edits))
    else:
        umod = icu.UnicodeString(op(icu.Locale(locale), ucur, edits))

    for is_change, old_len, new_len, old_i, new_i, _ in edits.getFineIterator():
        old_len = ucur.countChar32(old_i, old_len)
        if is_change:
            repl = str(umod[new_i:new_i+new_len])
            builder.replace(old_len, repl)
        else:
            builder.skip(old_len)

    return builder.build()


def casefold(bs: bistr) -> bistr:
    return _edit(bs, icu.CaseMap.fold)


def lower(bs: bistr, locale: Optional[str]) -> bistr:
    return _edit(bs, icu.CaseMap.toLower, locale)


def upper(bs: bistr, locale: Optional[str]) -> bistr:
    return _edit(bs, icu.CaseMap.toUpper, locale)


def title(bs: bistr, locale: Optional[str]) -> bistr:
    return _edit(bs, icu.CaseMap.toTitle, locale)


def _normalize(bs: bistr, normalizer: icu.Normalizer2) -> bistr:
    builder = BistrBuilder(bs)
    us = icu.UnicodeString(bs.modified)
    offset = 0
    while not builder.is_complete:
        i = normalizer.spanQuickCheckYes(us)
        builder.skip(us.countChar32(0, i))
        if builder.is_complete:
            break
        us = us[i:]

        i = 0
        while i < len(us):
            if us.charAt(i) & 0xFC00 == 0xD800:
                i += 1
            i += 1
            if normalizer.hasBoundaryBefore(chr(us.char32At(i))):
                break

        chunk = us[:i]
        normalized = str(normalizer.normalize(chunk))
        builder.replace(chunk.countChar32(), normalized)
        us = us[i:]

    return builder.build()


_NORMALIZERS = {
    'NFC': icu.Normalizer2.getNFCInstance,
    'NFKC': icu.Normalizer2.getNFKCInstance,
    'NFD': icu.Normalizer2.getNFDInstance,
    'NFKD': icu.Normalizer2.getNFKDInstance,
}

def normalize(bs: bistr, form: str) -> bistr:
    factory = _NORMALIZERS.get(form)
    if factory:
        return _normalize(bs, factory())
    else:
        raise ValueError('invalid normalization form')



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

from typing import Callable, Match, Pattern, Tuple, Union


Bounds = Tuple[int, int]

Range = Union[range, slice, Bounds]

Regex = Union[str, Pattern[str]]

Replacement = Union[str, Callable[[Match[str]], str]]

String = Union[str, 'bistr']



#!/usr/bin/env python3

# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT license.

from setuptools import setup


setup(
    name='bistring',
    version='0.1',
    author='Microsoft Research Montreal',
    author_email='msrmtle@microsoft.com',
    description='Bidirectionally transformed strings',
    url='https://github.com/microsoft/bistring',
    packages=[
        'bistring',
    ],
    test_suite='tests',
    setup_requires=[
        'pytest-runner',
    ],
    install_requires=[
        'pyicu',
    ],
    extras_require={
        'test': [
            'pytest',
            'regex',
        ],
    },
    tests_require=[
        'bistring[test]',
    ],
)
